from typing import List

import PIL.Image
import numpy as np
import pytest

from instance_segmentation import DataStorage, SegmentationObject
from instance_segmentation.model import SegmentationModelBase


@pytest.fixture(scope="session")
def storage():
    storage = DataStorage()
    yield storage
    storage.clean()


test_object = SegmentationObject(class_i=1, x=0, y=0, w=1, h=1, mask=[[]])


class SegmentationModelTest(SegmentationModelBase):
    def __init__(self, storage: DataStorage):
        super().__init__(classes=["fg", "bg"], model_path='', storage=storage)

    def _load_model(self, model_path: str):
        pass

    def predict(self, input: np.ndarray) -> List[SegmentationObject]:
        return [test_object]


def test_segmentation_model_base_call(storage: DataStorage):
    model = SegmentationModelTest(storage)

    job_id = "abc"
    array = np.array([[0, 0], [1, 1], [2, 2]])
    image = PIL.Image.fromarray(array.astype(np.uint8))
    storage.save(job_id, image, output=False)

    model(job_id)

    objects = storage.load(job_id, output=True)
    assert objects == [test_object]



import pickle
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import List, Any


class DataStorage:

    def __init__(self, base_path: str = "./data/storage"):
        self._base_path = Path(base_path)
        self._base_path.mkdir(parents=True, exist_ok=True)

    def clean(self):
        shutil.rmtree(self._base_path, ignore_errors=True)
        self._base_path.mkdir(parents=True, exist_ok=True)

    def _data_path(self, gid: str, output: bool) -> Path:
        type_ = "output" if output else "input"
        return self._base_path / f"{gid}-{type_}.pickle"

    def save(self, gid: str, obj: Any, output: bool):
        pickle.dump(obj, self._data_path(gid, output).open("wb"))

    def load(self, gid: str, output: bool) -> Any:
        image_path = self._data_path(gid, output)
        if image_path.exists():
            return pickle.load(image_path.open("rb"))


@dataclass
class SegmentationObject:
    """Instance segmentation object.

    Attributes:
        class_i: class index
        x: X coordinate of top left corner
        y: Y coordinate of top left corner
        w: object width
        h: object height
        mask: 2D mask of object pixels
    """

    class_i: int
    x: int
    y: int
    w: int
    h: int
    mask: List[List[int]]



import falcon
from falcon.http_status import HTTPStatus

from telesto.logger import logger
from telesto.config import config
from telesto.models import ModelType


class HandleCORS(object):
    def process_request(self, req, resp):
        resp.set_header("Access-Control-Allow-Origin", "*")
        resp.set_header("Access-Control-Allow-Methods", "*")
        resp.set_header("Access-Control-Allow-Headers", "*")
        resp.set_header("Access-Control-Max-Age", 1728000)  # 20 days
        if req.method == "OPTIONS":
            raise HTTPStatus(falcon.HTTP_200, body="\n")


class AuthMiddleware(object):

    @staticmethod
    def _api_key_is_valid(auth):
        try:
            _, api_key = auth.split(" ")
            assert config.get("common", "api_key") == api_key, "API key mismatch"
        except Exception as e:
            logger.warning(f"Auth error: {e}")
            return False

        return True

    def process_request(self, req, resp):
        challenges = ["Bearer realm='Access to protected endpoint'"]
        auth = req.get_header("Authorization")

        if auth is None:
            description = "Please provide an API key as part of the request."
            raise falcon.HTTPUnauthorized("API key required", description, challenges)

        if not self._api_key_is_valid(auth):
            description = (
                "The provided API key is not valid. Please request a API key and try again."
            )
            raise falcon.HTTPUnauthorized("Authentication required", description, challenges)


def get_app():
    middleware = [HandleCORS()]
    if config.get("common", "api_key"):
        middleware.append(AuthMiddleware())
    api = falcon.API(middleware=middleware)

    model_type = config.get("common", "model_type")
    if model_type == ModelType.INSTANCE_SEGMENTATION:
        from telesto.instance_segmentation.app import add_routes
    elif model_type == ModelType.CLASSIFICATION:
        from telesto.classification.app import add_routes
    else:
        raise Exception(f"Wrong model type: {model_type}")

    add_routes(api)
    return api


if __name__ == "__main__":
    logger.info("Starting dev API server...")

    from wsgiref import simple_server

    httpd = simple_server.make_server("0.0.0.0", 9876, get_app())
    logger.info("Dev API server started")
    httpd.serve_forever()



import base64
import io
import os
import json
import socket
from importlib import import_module
from typing import List

import falcon
import numpy as np
import PIL.Image

from telesto.logger import logger
from telesto.models import RandomClassificationModel


def preprocess(doc: dict) -> List[np.ndarray]:
    input_list = []
    for image_doc in doc["images"]:
        image_bytes = base64.b64decode(image_doc["content"])
        image = PIL.Image.open(io.BytesIO(image_bytes))
        array = np.array(image)
        assert array.ndim in [2, 3], f"Wrong number of dimensions: {array.ndim}"

        input_list.append(array)
    return input_list


def postprocess(pred_array: np.ndarray, classes: List[str]) -> dict:
    predictions = []
    for pred in pred_array:
        class_probs = {classes[i]: round(float(prob), 5) for i, prob in enumerate(pred)}
        class_prediction = classes[pred.argmax()]
        predictions.append({"probs": class_probs, "prediction": class_prediction})
    return {"predictions": predictions}


class ClassificationResource:
    def __init__(self):
        try:
            self.model_wrapper = self._load_model()
        except ModuleNotFoundError as e:
            if int(os.environ.get("USE_FALLBACK_MODEL", 0)):
                logger.warning(
                    "No model module found. Using fallback model 'RandomClassificationModel'"
                )
                self.model_wrapper = RandomClassificationModel()
            else:
                raise e

    @staticmethod
    def _load_model():
        module = import_module("model")
        model_class = getattr(module, "ClassificationModel")
        return model_class()

    def on_get(self, req: falcon.Request, resp: falcon.Response):
        doc = {"status": "ok", "host": socket.getfqdn(), "worker.pid": os.getpid()}
        resp.body = json.dumps(doc, ensure_ascii=False)

    def on_post(self, req: falcon.Request, resp: falcon.Response):
        try:
            req_doc = json.load(req.bounded_stream)
            input_list = preprocess(req_doc)
            pred_array = self.model_wrapper(input_list)
            resp_doc = postprocess(pred_array, self.model_wrapper.classes)
            resp.body = json.dumps(resp_doc)
        except ValueError as e:
            raise falcon.HTTPError(falcon.HTTP_400, description=str(e))
        except Exception as e:
            logger.error(e, exc_info=True)
            raise falcon.HTTPError(falcon.HTTP_500)


def add_routes(api: falcon.API):
    api.add_route("/", ClassificationResource())



import base64
import io
import os
import socket
import threading
import time
from dataclasses import asdict
from importlib import import_module
from typing import Deque, List
from uuid import uuid4
import json
from collections import deque

import PIL
from PIL.Image import Image
import falcon

from telesto.logger import logger
from telesto.config import config
from telesto.instance_segmentation import SegmentationObject, DataStorage
from telesto.instance_segmentation.model import DummySegmentationModel, SegmentationModelBase


INPUT_IMAGE_FORMAT = {
    "type": "png",
    "palette": "RGB24",
    "encoding": "base64",
    "max_size": "5120",
}

OUTPUT_OBJECT_MASK_FORMAT = {
    "type": "json",
    "palette": "GREY8",
    "encoding": "plain",
}

API_DOCS = {
    "name": config.get("common", "name"),
    "description": config.get("common", "desc"),
    "authentication": {
        "header": "Authorization",
        "schema": "Bearer <API_KEY>"
    },
    "endpoints": [
        {
            "path": "/",
            "method": "GET",
            "name": "Status endpoint",
            "description": "Returns status of the API",
        },
        {
            "path": "/docs",
            "method": "GET",
            "name": "Documentation endpoint",
            "description": "Returns this information",
        },
        {
            "path": "/jobs/",
            "method": "POST",
            "request_body": {
                "image": "<str>",
            },
            "image_format": {
                **INPUT_IMAGE_FORMAT,
            },
            "response_body": {
                "job_id": "<UUID>"
            }
        },
        {
            "path": "/jobs/<job_id>",
            "method": "GET",
            "response_body": {
                "objects": [
                    {
                        "class_i": "<int>",
                        "x": "<int>",
                        "y": "<int>",
                        "w": "<int>",
                        "h": "<int>",
                        "mask": "<str>"
                    }
                ],
            },
            "object_mask_format": {
                **OUTPUT_OBJECT_MASK_FORMAT,
            },
            "classes": config.get("common", "classes").split(","),
        }
    ]
}


def preprocess(doc: dict) -> Image:
    try:
        image_bytes = base64.b64decode(doc["image"])
        image = PIL.Image.open(io.BytesIO(image_bytes))
        assert image.mode == "RGB", f"Wrong image mode: {image.mode} != 'RGB'"
    except Exception as e:
        raise ValueError(e)
    return image


def postprocess(objects: List[SegmentationObject]) -> dict:
    return {"objects": [asdict(obj) for obj in objects]}


class SegmentationBase:
    def on_get(self, req, resp):
        doc = {"status": "ok", "host": socket.getfqdn(), "worker.pid": os.getpid()}
        resp.body = json.dumps(doc, ensure_ascii=False)


class SegmentationDocs:
    def on_get(self, req, resp):
        resp.body = json.dumps(API_DOCS, ensure_ascii=False)


class SegmentationJobs:

    def __init__(self, storage: DataStorage, job_queue: Deque):
        self._storage = storage
        self._job_queue = job_queue

    def on_post(self, req: falcon.Request, resp: falcon.Response):
        try:
            req_doc = json.load(req.bounded_stream)
            assert "image" in req_doc, f"'image' not found in {req_doc}"

            job_id = uuid4().hex
            image = preprocess(req_doc)
            self._storage.save(job_id, image, output=False)
            self._job_queue.appendleft(job_id)

            resp.status = falcon.HTTP_CREATED
            resp.body = json.dumps({"job_id": job_id})
        except (ValueError, AssertionError) as e:
            raise falcon.HTTPError(falcon.HTTP_400, description=str(e))
        except Exception as e:
            logger.error(e, exc_info=True)
            raise falcon.HTTPError(falcon.HTTP_500)


class SegmentationJob:

    def __init__(self, storage: DataStorage):
        self._storage = storage

    def on_get(self, req: falcon.Request, resp: falcon.Response, job_id: str):
        try:
            objects = self._storage.load(job_id, output=True)
            assert objects is not None, "No data found"

            resp_doc = postprocess(objects)
            resp.body = json.dumps(resp_doc)
        except AssertionError as e:
            raise falcon.HTTPError(falcon.HTTP_404, description=str(e))
        except ValueError as e:
            raise falcon.HTTPError(falcon.HTTP_400, description=str(e))
        except Exception as e:
            logger.error(e, exc_info=True)
            raise falcon.HTTPError(falcon.HTTP_500)


def load_model(storage: DataStorage) -> SegmentationModelBase:
    try:
        module = import_module("model")
        model_class = getattr(module, "SegmentationModel")
        return model_class(storage)
    except ModuleNotFoundError as e:
        if int(os.environ.get("USE_FALLBACK_MODEL", 0)):
            logger.warning(
                "No 'model' module found. Using fallback model 'DummySegmentationModel'"
            )
            return DummySegmentationModel(storage)
        else:
            raise e


def start_worker(storage: DataStorage, job_queue: Deque):

    def thread_function():
        logger.info("Starting worker thread")
        model_wrapper = load_model(storage)
        logger.info("Worker thread started")

        while True:
            if job_queue:
                job_id = job_queue.pop()
                logger.info(f"Processing task {job_id}")
                model_wrapper(job_id)
                logger.info(f"Finished task {job_id}")
            else:
                time.sleep(1)

    thread = threading.Thread(target=thread_function, args=(), daemon=True)
    thread.start()


def add_routes(api: falcon.API):
    storage = DataStorage()
    job_queue = deque()
    start_worker(storage, job_queue)

    api.add_route("/", SegmentationBase())
    api.add_route("/docs", SegmentationDocs())
    # Note: falcon internally strips trailing slashes from requests
    api.add_route("/jobs", SegmentationJobs(storage, job_queue))
    api.add_route("/jobs/{job_id}", SegmentationJob(storage))



import logging

logger = logging.getLogger("telesto")
gunicorn_logger = logging.getLogger("gunicorn")
if gunicorn_logger.handlers:
    logger.handlers = gunicorn_logger.handlers
    logger.setLevel(gunicorn_logger.level)
else:
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    logger.addHandler(ch)



from enum import Enum

from telesto.classification.model import ClassificationModelBase, RandomClassificationModel
from telesto.instance_segmentation.model import SegmentationModelBase, DummySegmentationModel


class ModelType(str, Enum):
    CLASSIFICATION = "CLASSIFICATION"
    INSTANCE_SEGMENTATION = "INSTANCE_SEGMENTATION"



from __future__ import annotations
from typing import List

import numpy as np

from telesto.instance_segmentation import DataStorage, SegmentationObject


class SegmentationModelBase:
    """Base class for an instance segmentation model wrapper.

    Attributes:
        classes (list): contains the labels
        model: the object representing the model, to be loaded with _load_model()
    """

    def __init__(self, classes: List[str], model_path: str, storage: DataStorage):
        self._storage = storage
        self.model = self._load_model(model_path=model_path)
        self.classes = classes

    def _load_model(self, model_path: str) -> object:
        """Load model weights from path, create and return a model object."""

        raise NotImplemented

    def predict(self, input: np.ndarray) -> List[SegmentationObject]:
        """Segment input image and return a list of found objects.

        Args:
            input: input image, 3D array with 1 or 3 image channels

        Returns:
            list of found objects
        """
        raise NotImplemented

    def __call__(self, job_id: str):
        image = self._storage.load(job_id, output=False)
        objects = self.predict(np.asarray(image))
        self._storage.save(job_id, objects, output=True)


class DummySegmentationModel(SegmentationModelBase):
    def __init__(self, storage: DataStorage):
        super().__init__(classes=["fg", "bg"], model_path="", storage=storage)

    def _load_model(self, model_path: str):
        pass

    def predict(self, input: np.ndarray) -> List[SegmentationObject]:
        return [SegmentationObject(class_i=1, x=10, y=10, w=2, h=2, mask=[[0, 1], [0, 1]])]



from typing import List
import random

import numpy as np


class ClassificationModelBase:
    """Base class for a classification model wrapper.

    Attributes:
        classes (list): contains the labels
        model: the object representing the model, to be loaded with _load_model()
    """

    def __init__(self, classes: List[str], model_path: str):
        self.classes: List[str] = classes
        self.model = self._load_model(model_path=model_path)

    def _load_model(self, model_path: str) -> object:
        """Load model weights from path, create and return a model object."""

        raise NotImplemented

    def predict(self, input_list: List[np.ndarray]) -> np.ndarray:
        """Classify a list of input images and return an array of class probabilities.

        Args:
            input_list: list of input images, each image is a 3D array
                with a number of channels

        Returns:
            2D array of class probabilities, 0 dim - images, 1 dim - classes
        """
        raise NotImplemented

    def __call__(self, input_list: List[np.ndarray]) -> np.ndarray:
        if not (0 < len(input_list) <= 32):
            raise ValueError(f"Wrong number of images: {len(input_list)}")

        return self.predict(input_list)


class RandomClassificationModel(ClassificationModelBase):
    def __init__(self):
        super().__init__(classes=["cat", "dog"], model_path='')

    def _load_model(self, model_path: str):
        pass

    def predict(self, input_list: List[np.ndarray]) -> np.ndarray:
        batch_size = len(input_list)
        predictions = []
        for _ in range(batch_size):
            p = random.random()
            predictions.append([p, 1 - p])
        return np.array(predictions)



import os
import base64
import io
import json
import time

import falcon
from falcon import testing
import pytest
import numpy as np
import PIL.Image

from telesto.app import get_app
from telesto.config import config
from telesto.models import ModelType
from telesto.instance_segmentation import DataStorage, SegmentationObject

os.environ["USE_FALLBACK_MODEL"] = "1"


@pytest.fixture(scope="session", autouse=True)
def config_fixture():
    config["common"]["model_type"] = ModelType.INSTANCE_SEGMENTATION.value
    config["common"]["api_key"] = ""


@pytest.fixture
def client():
    return testing.TestClient(get_app())


@pytest.fixture(scope="session")
def storage():
    storage = DataStorage()
    yield storage
    storage.clean()


def make_test_image(rgb=False):
    array = np.array([[0, 0], [1, 1], [2, 2]])
    if rgb:
        array = np.stack([array] * 3, axis=2)
    image = PIL.Image.fromarray(array.astype(np.uint8))
    return image


def test_image_storage_save_load(storage: DataStorage):
    image = make_test_image()

    gid = "abc"
    storage.save(gid, image, output=False)
    restored_image = storage.load(gid, output=False)

    assert np.all(np.asarray(image) == np.asarray(restored_image))


def test_image_storage_not_found(storage: DataStorage):
    restored_image = storage.load(gid="xyz", output=False)

    assert restored_image is None


def test_root_get(client: testing.TestClient):
    resp = client.simulate_get("/")

    assert resp.status == falcon.HTTP_OK

    resp_doc = json.loads(resp.content)
    assert resp_doc["status"] == "ok"


def test_docs_get(client: testing.TestClient):
    resp = client.simulate_get("/docs")

    assert resp.status == falcon.HTTP_OK

    resp_doc = json.loads(resp.content)
    assert "endpoints" in resp_doc and resp_doc["endpoints"]


def test_segm_get(client: testing.TestClient, storage: DataStorage):
    job_id = "bcd"
    test_segm_object = SegmentationObject(class_i=1, x=0, y=0, w=1, h=1, mask=[[]])
    storage.save(job_id, [test_segm_object], output=True)

    resp = client.simulate_get(f"/jobs/{job_id}")

    assert resp.status == falcon.HTTP_OK

    resp_doc = json.loads(resp.content)
    assert SegmentationObject(**resp_doc["objects"][0]) == test_segm_object


def test_segm_post_get(client: testing.TestClient):
    image = make_test_image(rgb=True)
    fp = io.BytesIO()
    image.save(fp, format="PNG")
    fp.seek(0)
    req_doc = {"image": base64.b64encode(fp.read()).decode()}

    resp = client.simulate_post(
        "/jobs", body=json.dumps(req_doc), headers={"content-type": "application/json"}
    )

    assert resp.status == falcon.HTTP_CREATED, resp.text

    resp_doc = json.loads(resp.content)
    assert resp_doc["job_id"], resp_doc

    resp = client.simulate_get(f"/jobs/{resp_doc['job_id']}")
    assert resp.status == falcon.HTTP_404, resp.text

    time.sleep(1)

    resp = client.simulate_get(f"/jobs/{resp_doc['job_id']}")
    assert resp.status == falcon.HTTP_200, resp.text

    resp_doc = json.loads(resp.content)
    assert "objects" in resp_doc, resp_doc


def test_api_key_auth_error():
    config["common"]["api_key"] = "API_KEY"

    client = testing.TestClient(get_app())
    resp = client.simulate_get("/")

    assert resp.status == falcon.HTTP_401


def test_api_key_auth_ok():
    config["common"]["api_key"] = "API_KEY"

    client = testing.TestClient(get_app())
    resp = client.simulate_get("/", headers={"Authorization": "Bearer API_KEY"})

    assert resp.status == falcon.HTTP_OK
